import json
import os

import pytest

import emod_api.campaign as camp
import emod_api.interventions.outbreak as ob

from idmtools.core import ItemType
from idmtools.core.platform_factory import Platform
from idmtools.entities.experiment import Experiment
from idmtools_test.utils.itest_with_persistence import ITestWithPersistence

from emodpy.emod_task import EMODTask
from emodpy.reporters.custom import Report_TBHIV_ByAge
from examples.config_update_parameters import del_folder
from examples.serialization.globals import config_update_params

from . import manifest

current_directory = os.path.dirname(os.path.realpath(__file__))
INPUT_PATH = os.path.join(current_directory, "..", "examples", "reporters", "inputs")
BIN_PATH = os.path.join(current_directory, "..", "examples", "inputs", "bin")


@pytest.mark.emod
class TestTBHIVReports(ITestWithPersistence):
    def setUp(self) -> None:
        self.case_name = os.path.basename(__file__) + "--" + self._testMethodName
        print(self.case_name)
        self.platform = Platform('SLURM')
        self.eradication_path = manifest.eradication_path_tbhiv_linux
        self.schema_path = manifest.schema_path_tbhiv_linux

    def test_custom_report_Report_TBHIV_ByAge(self):
        def build_camp():
            """
            Build a campaign input file for the DTK using emod_api.
            """
            seed = ob.seed_by_coverage(1, camp, 0.05)
            camp.add(seed, first=True)
            return camp

        def set_param_fn(config):
            config.parameters.Simulation_Type = "TBHIV_SIM"
            config.parameters.TB_Active_Period_Distribution = 'GAUSSIAN_DISTRIBUTION'
            config.parameters.Base_Infectivity_Constant = 0.2
            config.parameters.Base_Mortality = 0
            config.parameters.Simulation_Duration = 365
            # config.parameters.Event_Map = {
            #   "GP_EVENT_000": "Hello",
            #   "GP_EVENT_001": "Oh_No_I_Have_HIV",
            #   "GP_EVENT_002": "Yay"
            # }
            return config

        camp.schema_path = self.schema_path

        report = Report_TBHIV_ByAge()
        report.add_report(200, 0, 0, 200)
        # report.add_report(100, 0, 0, 100)

        report.asset_dir = manifest.plugins_folder_tbhiv

        task = EMODTask.from_default2(
            config_path="my_config.json",
            eradication_path=self.eradication_path,
            campaign_builder=build_camp,
            schema_path=self.schema_path,
            param_custom_cb=set_param_fn,
            ep4_custom_cb=None,
            demog_builder=None,
            plugin_report=report
        )

        self.assertTrue(len(task.reporters.assets) > 0)
        self.assertTrue(len(task.reporters.custom_reporters) == 1)
        self.assertTrue(len(task.reporters.custom_reporters[0].Reports) == 1)

        # Create the experiment from task and run
        experiment = Experiment.from_task(task, name=self.case_name)
        self.platform.run_items(experiment)
        self.platform.wait_till_done(experiment)

        # skip this assertion since the Eradication crash at the end of the simulation: https://github.com/InstituteforDiseaseModeling/DtkTrunk/issues/4274
        # self.assertTrue(experiment.succeeded, "expected experiment succeeded")

        # verify:
        # 1) download output/Report_TBHIV_ByAge.csv from simulation, verify download file exists
        # 2) download 'custom_reports.json' from simulation working dir, validate it has correct content

        # delete output from previous run
        del_folder(os.path.join(manifest.current_directory, "output_tbhiv"))
        # create a new empty 'output' dir
        os.mkdir(os.path.join(manifest.current_directory, "output_tbhiv"))
        filenames = ["output/Report_TBHIV_ByAge.csv", "custom_reports.json"]

        sims = self.platform.get_children_by_object(experiment)
        for simulation in sims:
            # download files from simulation
            self.platform.get_files_by_id(simulation.id, item_type=ItemType.SIMULATION, files=filenames,
                                          output=os.path.join(manifest.current_directory, "output_tbhiv"))
            # validate files exist
            local_path = os.path.join(manifest.current_directory, "output_tbhiv", str(simulation.uid))
            self.assertTrue(os.path.exists(os.path.join(local_path, "output", "Report_TBHIV_ByAge.csv")))
            custom_reports_file_path = os.path.join(local_path, "custom_reports.json")
            self.assertTrue(os.path.exists(custom_reports_file_path))

            with open(custom_reports_file_path, "r") as f:
                custom_reports = json.load(f)
                print(custom_reports)
                reports = custom_reports["Custom_Reports"]["Report_TBHIV_ByAge"]["Reports"]
                self.assertDictEqual(reports[0],
                                     {'Additional_Events': [],
                                      'Max_Age_Yrs': 200,
                                      'Min_Age_Yrs': 0,
                                      'Start_Year': 0,
                                      'Stop_Year': 200})
                # self.assertDictEqual(reports[1],
                #                      {'Max_Age_Yrs': 100, 'Min_Age_Yrs': 0, 'Start_Year': 0, 'Stop_Year': 100})

    # Todo: skip for now, will need to update this test
    def _test_multiple_custom_reports(self):
        input = os.path.abspath(os.path.join(current_directory, "..", "examples", "serialization"))
        task = EMODTask.from_files(eradication_path=os.path.join(BIN_PATH, "Eradication.exe"),
                                   config_path=os.path.join(input, "inputs", "config.json"),
                                   campaign_path=os.path.join(input, "inputs", "campaign.json"),
                                   demographics_paths=os.path.join(input, "inputs", "3nodes_demographics.json"))

        config_update_params(task)

        # Add the DLLs to the collection
        task.reporters.add_dll_folder(os.path.join(input, "custom_reports", "reporter_plugins", "Windows"))
        task.reporters.read_custom_reports_file(os.path.join(input, "custom_reports", "custom_reports.json"))
        self.assertTrue(task.reporters.custom_reporters.count, 2)
        self.assertTrue(task.reporters.custom_reporters[0].name, 'ReportHumanMigrationTracking')
        self.assertTrue(task.reporters.custom_reporters[1].name, 'ReportNodeDemographics')

        experiment = Experiment.from_task(task, name=self.case_name)
        self.platform.run_items(experiment)
        self.platform.wait_till_done(experiment)
        self.assertTrue(experiment.succeeded, "expected experiment succeeded")

        # verify 2 custom reports in output of simulation
        # 1) download output/ReportHumanMigrationTracking.csv from simulation, verify file exists
        # 2) download output/ReportNodeDemographics.csv from simulation, validate file exists
        # 3) download output/DemographicsSummary.json from simulation, verify file exists

        # delete output from previous run
        del_folder(os.path.join(manifest.current_directory, 'output_tbhiv'))
        # create a new empty 'output' dir
        os.mkdir(os.path.join(manifest.current_directory, 'output_tbhiv'))
        filenames = ["output/ReportHumanMigrationTracking.csv", "output/ReportNodeDemographics.csv",
                     "output/DemographicsSummary.json"]

        sims = self.platform.get_children_by_object(experiment)
        for simulation in sims:
            # download files from simulation
            self.platform.get_files_by_id(simulation.id, item_type=ItemType.SIMULATION, files=filenames,
                                          output=os.path.join(manifest.current_directory, 'output_tbhiv'))
            # validate files exist
            local_path = os.path.join(manifest.current_directory, "output_tbhiv", str(simulation.uid), "output")
            self.assertTrue(os.path.exists(os.path.join(local_path, "ReportHumanMigrationTracking.csv")))
            self.assertTrue(os.path.exists(os.path.join(local_path, "ReportNodeDemographics.csv")))
            self.assertTrue(os.path.exists(os.path.join(local_path, "DemographicsSummary.json")))
